/*
 * Copyright (C) 2009 Splitted-Desktop Systems. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sub license, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.
 * IN NO EVENT SHALL PRECISION INSIGHT AND/OR ITS SUPPLIERS BE LIABLE FOR
 * ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include "config.h"
#include "va.h"
#include "va_compat.h"
#include "va_backend.h"
#include <string.h>
#include <alloca.h>
#include <assert.h>

#define CTX(dpy)        (((VADisplayContextP)dpy)->pDriverContext)
#define ASSERT		assert

/* Invoke the normal VA API entry-point. This one can be replaced with
   one of the va_compat_*() functions */
#define VA_INVOKE(FUNC, CTX, ...) \
    (CTX)->vtable.FUNC(CTX, __VA_ARGS__)

/* Invoke the real VA API entry-point. That is, the driver's function */
#define VA_INVOKE_COMPAT(FUNC, CTX, ...) \
    ((VACompatContextP)(CTX)->compat)->vtable.FUNC(CTX, __VA_ARGS__)

typedef struct VACompatContext *VACompatContextP;

typedef VAStatus (*VABufferTranslateFunc)(VACompatContextP ctx, void *dest, const void *src);

typedef enum {
    VA_DRIVER_ID_UNKNOWN,
    VA_DRIVER_ID_POULSBO,
    VA_DRIVER_ID_IEGD
} VADriverID;

typedef struct {
    VABufferID                  id;
    VABufferType                type;
    unsigned int                size;
    unsigned int                num_elements;
    unsigned char              *data;
    VABufferTranslateFunc       translate;
    VABufferID                  compat_id;
    unsigned int                compat_size;
    unsigned char              *compat_data;
    unsigned int                map_count;
} VABufferCompat;

typedef struct VACompatContext {
    VABufferCompat             *buffers;
    unsigned int                buffers_count_max;
    struct VADriverVTable       vtable;                 /* original vtable */
    VADriverID                  driver_id;
    char                       *driver_name;
    unsigned int                skip_frame : 1;
} VACompatContext;

#define COPY_ARRAY(DST, SRC, ARRAY) \
    memcpy((DST)->ARRAY, (SRC)->ARRAY, sizeof((DST)->ARRAY))
#define COPY_FIELD(DST, SRC, MEMBER) \
    (DST)->MEMBER = (SRC)->MEMBER
#define COPY_BIT_FIELD(DST, SRC, FIELD, MEMBER)  \
    (DST)->MEMBER = (SRC)->FIELD.bits.MEMBER
#define COPY_BIT_FLAG(DST, SRC, FIELD, MEMBER)  \
    (DST)->MEMBER = (SRC)->FIELD.flags.MEMBER

static VAStatus va_compat_TranslateVAPictureH264(
    VACompatContextP    ctx,
    void               *dest_pic,
    const void         *src_pic
)
{
    VAPictureH264_0_29 *dest = dest_pic;
    const VAPictureH264 *src = src_pic;

    COPY_FIELD(dest, src, picture_id);
    COPY_FIELD(dest, src, flags);
    COPY_FIELD(dest, src, TopFieldOrderCnt);
    COPY_FIELD(dest, src, BottomFieldOrderCnt);

    return VA_STATUS_SUCCESS;
}

static VAStatus va_compat_TranslateVAPictureParameterBufferH264(
    VACompatContextP    ctx,
    void               *dest_buffer,
    const void         *src_buffer
)
{
    VAPictureParameterBufferH264_0_29 *dest = dest_buffer;
    const VAPictureParameterBufferH264 *src = src_buffer;
    VAStatus status;
    int i;

    status = va_compat_TranslateVAPictureH264(ctx,
                                              &dest->CurrPic, &src->CurrPic);
    if (status != VA_STATUS_SUCCESS)
        return status;

    for (i = 0; i < 16; i++)
    {
        status = va_compat_TranslateVAPictureH264(ctx,
                                                  &dest->ReferenceFrames[i],
                                                  &src->ReferenceFrames[i]);
        if (status != VA_STATUS_SUCCESS)
            return status;
    }

    COPY_FIELD(dest, src, picture_width_in_mbs_minus1);
    COPY_FIELD(dest, src, picture_height_in_mbs_minus1);
    COPY_FIELD(dest, src, bit_depth_luma_minus8);
    COPY_FIELD(dest, src, bit_depth_chroma_minus8);
    COPY_FIELD(dest, src, num_ref_frames);
    dest->seq_fields = 0; /* reset all bits */
    COPY_BIT_FIELD(dest, src, seq_fields, chroma_format_idc);
    COPY_BIT_FIELD(dest, src, seq_fields, residual_colour_transform_flag);
    COPY_BIT_FIELD(dest, src, seq_fields, frame_mbs_only_flag);
    COPY_BIT_FIELD(dest, src, seq_fields, mb_adaptive_frame_field_flag);
    COPY_BIT_FIELD(dest, src, seq_fields, direct_8x8_inference_flag);
    COPY_BIT_FIELD(dest, src, seq_fields, MinLumaBiPredSize8x8);
    COPY_FIELD(dest, src, num_slice_groups_minus1);
    COPY_FIELD(dest, src, slice_group_map_type);
    COPY_FIELD(dest, src, pic_init_qp_minus26);
    COPY_FIELD(dest, src, chroma_qp_index_offset);
    COPY_FIELD(dest, src, second_chroma_qp_index_offset);
    dest->pic_fields = 0; /* reset all bits */
    COPY_BIT_FIELD(dest, src, pic_fields, entropy_coding_mode_flag);
    COPY_BIT_FIELD(dest, src, pic_fields, weighted_pred_flag);
    COPY_BIT_FIELD(dest, src, pic_fields, weighted_bipred_idc);
    COPY_BIT_FIELD(dest, src, pic_fields, transform_8x8_mode_flag);
    COPY_BIT_FIELD(dest, src, pic_fields, field_pic_flag);
    COPY_BIT_FIELD(dest, src, pic_fields, constrained_intra_pred_flag);
    COPY_FIELD(dest, src, frame_num);

    return VA_STATUS_SUCCESS;
}

static VAStatus va_compat_TranslateVASliceParameterBufferH264(
    VACompatContextP    ctx,
    void               *dest_buffer,
    const void         *src_buffer
)
{
    VASliceParameterBufferH264_0_29 *dest = dest_buffer;
    const VASliceParameterBufferH264 *src = src_buffer;
    VAStatus status;
    int i;

    COPY_FIELD(dest, src, slice_data_size);
    COPY_FIELD(dest, src, slice_data_offset);
    COPY_FIELD(dest, src, slice_data_flag);
    COPY_FIELD(dest, src, slice_data_bit_offset);
    COPY_FIELD(dest, src, first_mb_in_slice);
    COPY_FIELD(dest, src, slice_type);
    COPY_FIELD(dest, src, direct_spatial_mv_pred_flag);
    COPY_FIELD(dest, src, num_ref_idx_l0_active_minus1);
    COPY_FIELD(dest, src, num_ref_idx_l1_active_minus1);
    COPY_FIELD(dest, src, cabac_init_idc);
    COPY_FIELD(dest, src, slice_qp_delta);
    COPY_FIELD(dest, src, disable_deblocking_filter_idc);
    COPY_FIELD(dest, src, slice_alpha_c0_offset_div2);
    COPY_FIELD(dest, src, slice_beta_offset_div2);
    for (i = 0; i < 32; i++) {
        status = va_compat_TranslateVAPictureH264(ctx,
                                                  &dest->RefPicList0[i],
                                                  &src->RefPicList0[i]);
        if (status != VA_STATUS_SUCCESS)
            return status;
        status = va_compat_TranslateVAPictureH264(ctx,
                                                  &dest->RefPicList1[i],
                                                  &src->RefPicList1[i]);
        if (status != VA_STATUS_SUCCESS)
            return status;
    }
    COPY_FIELD(dest, src, luma_log2_weight_denom);
    COPY_FIELD(dest, src, chroma_log2_weight_denom);
    COPY_FIELD(dest, src, luma_weight_l0_flag);
    COPY_ARRAY(dest, src, luma_weight_l0);
    COPY_ARRAY(dest, src, luma_offset_l0);
    COPY_FIELD(dest, src, chroma_weight_l0_flag);
    COPY_ARRAY(dest, src, chroma_weight_l0);
    COPY_ARRAY(dest, src, chroma_offset_l0);
    COPY_FIELD(dest, src, luma_weight_l1_flag);
    COPY_ARRAY(dest, src, luma_weight_l1);
    COPY_ARRAY(dest, src, luma_offset_l1);
    COPY_FIELD(dest, src, chroma_weight_l1_flag);
    COPY_ARRAY(dest, src, chroma_weight_l1);
    COPY_ARRAY(dest, src, chroma_offset_l1);

    return VA_STATUS_SUCCESS;
}

static VAStatus va_compat_TranslateVAPictureParameterBufferVC1(
    VACompatContextP    ctx,
    void               *dest_buffer,
    const void         *src_buffer
)
{
    VAPictureParameterBufferVC1_0_29 *dest = dest_buffer;
    const VAPictureParameterBufferVC1 *src = src_buffer;

    COPY_FIELD(dest, src, forward_reference_picture);
    COPY_FIELD(dest, src, backward_reference_picture);
    COPY_FIELD(dest, src, inloop_decoded_picture);
    dest->sequence_fields = 0; /* reset all bits */
    COPY_BIT_FIELD(dest, src, sequence_fields, interlace);
    COPY_BIT_FIELD(dest, src, sequence_fields, syncmarker);
    COPY_BIT_FIELD(dest, src, sequence_fields, overlap);
    COPY_FIELD(dest, src, coded_width);
    COPY_FIELD(dest, src, coded_height);
    COPY_BIT_FIELD(dest, src, entrypoint_fields, closed_entry);
    COPY_BIT_FIELD(dest, src, entrypoint_fields, broken_link);
    COPY_BIT_FIELD(dest, src, entrypoint_fields, loopfilter);
    COPY_FIELD(dest, src, conditional_overlap_flag);
    COPY_FIELD(dest, src, fast_uvmc_flag);
    dest->range_mapping_fields = 0; /* reset all bits */
    dest->range_mapping_luma_flag   = src->range_mapping_fields.bits.luma_flag;
    dest->range_mapping_luma        = src->range_mapping_fields.bits.luma;
    dest->range_mapping_chroma_flag = src->range_mapping_fields.bits.chroma_flag;
    dest->range_mapping_chroma      = src->range_mapping_fields.bits.chroma;
    COPY_FIELD(dest, src, b_picture_fraction);
    COPY_FIELD(dest, src, cbp_table);
    COPY_FIELD(dest, src, mb_mode_table);
    COPY_FIELD(dest, src, range_reduction_frame);
    COPY_FIELD(dest, src, rounding_control);
    COPY_FIELD(dest, src, post_processing);
    COPY_FIELD(dest, src, picture_resolution_index);
    COPY_FIELD(dest, src, luma_scale);
    COPY_FIELD(dest, src, luma_shift);
    dest->picture_fields = 0; /* reset all bits */
    COPY_BIT_FIELD(dest, src, picture_fields, picture_type);
    COPY_BIT_FIELD(dest, src, picture_fields, frame_coding_mode);
    COPY_BIT_FIELD(dest, src, picture_fields, top_field_first);
    COPY_BIT_FIELD(dest, src, picture_fields, is_first_field);
    COPY_BIT_FIELD(dest, src, picture_fields, intensity_compensation);
    dest->raw_coding_flag = 0; /* reset all bits */
    COPY_BIT_FLAG(dest, src, raw_coding, mv_type_mb);
    COPY_BIT_FLAG(dest, src, raw_coding, direct_mb);
    COPY_BIT_FLAG(dest, src, raw_coding, skip_mb);
    COPY_BIT_FLAG(dest, src, raw_coding, field_tx);
    COPY_BIT_FLAG(dest, src, raw_coding, forward_mb);
    COPY_BIT_FLAG(dest, src, raw_coding, ac_pred);
    COPY_BIT_FLAG(dest, src, raw_coding, overflags);
    dest->bitplane_present_flag = 0; /* reset all bits */
    COPY_BIT_FLAG(dest, src, bitplane_present, bp_mv_type_mb);
    COPY_BIT_FLAG(dest, src, bitplane_present, bp_direct_mb);
    COPY_BIT_FLAG(dest, src, bitplane_present, bp_skip_mb);
    COPY_BIT_FLAG(dest, src, bitplane_present, bp_field_tx);
    COPY_BIT_FLAG(dest, src, bitplane_present, bp_forward_mb);
    COPY_BIT_FLAG(dest, src, bitplane_present, bp_ac_pred);
    COPY_BIT_FLAG(dest, src, bitplane_present, bp_overflags);
    dest->reference_fields = 0; /* reset all bits */
    COPY_BIT_FIELD(dest, src, reference_fields, reference_distance_flag);
    COPY_BIT_FIELD(dest, src, reference_fields, reference_distance);
    COPY_BIT_FIELD(dest, src, reference_fields, num_reference_pictures);
    COPY_BIT_FIELD(dest, src, reference_fields, reference_field_pic_indicator);
    dest->mv_fields = 0; /* reset all bits */
    COPY_BIT_FIELD(dest, src, mv_fields, mv_mode);
    COPY_BIT_FIELD(dest, src, mv_fields, mv_mode2);
    COPY_BIT_FIELD(dest, src, mv_fields, mv_table);
    COPY_BIT_FIELD(dest, src, mv_fields, two_mv_block_pattern_table);
    COPY_BIT_FIELD(dest, src, mv_fields, four_mv_switch);
    COPY_BIT_FIELD(dest, src, mv_fields, four_mv_block_pattern_table);
    COPY_BIT_FIELD(dest, src, mv_fields, extended_mv_flag);
    COPY_BIT_FIELD(dest, src, mv_fields, extended_mv_range);
    COPY_BIT_FIELD(dest, src, mv_fields, extended_dmv_flag);
    COPY_BIT_FIELD(dest, src, mv_fields, extended_dmv_range);
    dest->pic_quantizer_fields = 0; /* reset all bits */
    COPY_BIT_FIELD(dest, src, pic_quantizer_fields, dquant);
    COPY_BIT_FIELD(dest, src, pic_quantizer_fields, quantizer);
    COPY_BIT_FIELD(dest, src, pic_quantizer_fields, half_qp);
    COPY_BIT_FIELD(dest, src, pic_quantizer_fields, pic_quantizer_scale);
    COPY_BIT_FIELD(dest, src, pic_quantizer_fields, pic_quantizer_type);
    COPY_BIT_FIELD(dest, src, pic_quantizer_fields, dq_frame);
    COPY_BIT_FIELD(dest, src, pic_quantizer_fields, dq_profile);
    COPY_BIT_FIELD(dest, src, pic_quantizer_fields, dq_sb_edge);
    COPY_BIT_FIELD(dest, src, pic_quantizer_fields, dq_db_edge);
    COPY_BIT_FIELD(dest, src, pic_quantizer_fields, dq_binary_level);
    COPY_BIT_FIELD(dest, src, pic_quantizer_fields, alt_pic_quantizer);
    dest->transform_fields = 0; /* reset all bits */
    COPY_BIT_FIELD(dest, src, transform_fields, variable_sized_transform_flag);
    COPY_BIT_FIELD(dest, src, transform_fields, mb_level_transform_type_flag);
    COPY_BIT_FIELD(dest, src, transform_fields, frame_level_transform_type);
    COPY_BIT_FIELD(dest, src, transform_fields, transform_ac_codingset_idx1);
    COPY_BIT_FIELD(dest, src, transform_fields, transform_ac_codingset_idx2);
    COPY_BIT_FIELD(dest, src, transform_fields, intra_transform_dc_table);

    if (src->picture_fields.bits.picture_type == 4) {
        dest->picture_type = 1; /* P-frame */
        ctx->skip_frame    = 1;
    }

    return VA_STATUS_SUCCESS;
}

static VAStatus va_compat_TranslateVAPictureParameterBufferMPEG2(
    VACompatContextP    ctx,
    void               *dest_buffer,
    const void         *src_buffer
)
{
    VAPictureParameterBufferMPEG2_0_29 *dest = dest_buffer;
    const VAPictureParameterBufferMPEG2 *src = src_buffer;

    COPY_FIELD(dest, src, horizontal_size);
    COPY_FIELD(dest, src, vertical_size);
    COPY_FIELD(dest, src, forward_reference_picture);
    COPY_FIELD(dest, src, backward_reference_picture);
    COPY_FIELD(dest, src, picture_coding_type);
    COPY_FIELD(dest, src, f_code);
    dest->picture_coding_extension = 0; /* reset all bits */
    COPY_BIT_FIELD(dest, src, picture_coding_extension, intra_dc_precision);
    COPY_BIT_FIELD(dest, src, picture_coding_extension, picture_structure);
    COPY_BIT_FIELD(dest, src, picture_coding_extension, top_field_first);
    COPY_BIT_FIELD(dest, src, picture_coding_extension, frame_pred_frame_dct);
    COPY_BIT_FIELD(dest, src, picture_coding_extension, concealment_motion_vectors);
    COPY_BIT_FIELD(dest, src, picture_coding_extension, q_scale_type);
    COPY_BIT_FIELD(dest, src, picture_coding_extension, intra_vlc_format);
    COPY_BIT_FIELD(dest, src, picture_coding_extension, alternate_scan);
    COPY_BIT_FIELD(dest, src, picture_coding_extension, repeat_first_field);
    COPY_BIT_FIELD(dest, src, picture_coding_extension, progressive_frame);
    COPY_BIT_FIELD(dest, src, picture_coding_extension, is_first_field);

    return VA_STATUS_SUCCESS;
}

static VAStatus va_compat_TranslateVAPictureParameterBufferMPEG4(
    VACompatContextP    ctx,
    void               *dest_buffer,
    const void         *src_buffer
)
{
    VAPictureParameterBufferMPEG4_0_29 *dest = dest_buffer;
    const VAPictureParameterBufferMPEG4 *src = src_buffer;

    COPY_FIELD(dest, src, vop_width);
    COPY_FIELD(dest, src, vop_height);
    COPY_FIELD(dest, src, forward_reference_picture);
    COPY_FIELD(dest, src, backward_reference_picture);
    dest->vol_fields = 0; /* reset all bits */
    COPY_BIT_FIELD(dest, src, vol_fields, short_video_header);
    COPY_BIT_FIELD(dest, src, vol_fields, chroma_format);
    COPY_BIT_FIELD(dest, src, vol_fields, interlaced);
    COPY_BIT_FIELD(dest, src, vol_fields, obmc_disable);
    COPY_BIT_FIELD(dest, src, vol_fields, sprite_enable);
    COPY_BIT_FIELD(dest, src, vol_fields, sprite_warping_accuracy);
    COPY_BIT_FIELD(dest, src, vol_fields, quant_type);
    COPY_BIT_FIELD(dest, src, vol_fields, quarter_sample);
    COPY_BIT_FIELD(dest, src, vol_fields, data_partitioned);
    COPY_BIT_FIELD(dest, src, vol_fields, reversible_vlc);
    COPY_FIELD(dest, src, no_of_sprite_warping_points);
    COPY_ARRAY(dest, src, sprite_trajectory_du);
    COPY_ARRAY(dest, src, sprite_trajectory_dv);
    COPY_FIELD(dest, src, quant_precision);
    dest->vop_fields = 0; /* reset all bits */
    COPY_BIT_FIELD(dest, src, vop_fields, vop_coding_type);
    COPY_BIT_FIELD(dest, src, vop_fields, backward_reference_vop_coding_type);
    COPY_BIT_FIELD(dest, src, vop_fields, vop_rounding_type);
    COPY_BIT_FIELD(dest, src, vop_fields, intra_dc_vlc_thr);
    COPY_BIT_FIELD(dest, src, vop_fields, top_field_first);
    COPY_BIT_FIELD(dest, src, vop_fields, alternate_vertical_scan_flag);
    COPY_FIELD(dest, src, vop_fcode_forward);
    COPY_FIELD(dest, src, vop_fcode_backward);
    COPY_FIELD(dest, src, num_gobs_in_vop);
    COPY_FIELD(dest, src, num_macroblocks_in_gob);
    COPY_FIELD(dest, src, TRB);
    COPY_FIELD(dest, src, TRD);

    return VA_STATUS_SUCCESS;
}

#undef COPY_BIT_FLAG
#undef COPY_BIT_FIELD
#undef COPY_FIELD
#undef COPY_ARRAY

static inline int va_IsIntelBuffer(VADriverContextP ctx, VABufferID id)
{
    VACompatContextP compat = ctx->compat;

    if (compat->driver_id == VA_DRIVER_ID_POULSBO) {
        /* There were only Intel implementations for VA API. Besides, the
           returned buffer IDs had the following format [ 0 BASE ID ] where
           BASE is the 7-bit value 0x04 and ID an index into the heap of
           objects */
        return (id & 0x7f000000) == 0x04000000;
    }

    if (compat->driver_id == VA_DRIVER_ID_IEGD) {
        /* XXX: there is no means to differentiate the buffers, they
           are linearly generated (IEGD 10.0 build 1335) */
        return 1;
    }

    return 0;
}

static VABufferCompat *va_GetBufferCompat (
    VADriverContextP ctx,
    VABufferID id
)
{
    VACompatContextP compat = ctx->compat;
    int index;

    if (!va_IsIntelBuffer(ctx, id))
        return NULL;

    index = id & 0x00ffffff;
    if (index >= compat->buffers_count_max)
        return NULL;

    if (compat->buffers[index].id != id)
        return NULL;

    return &compat->buffers[index];
}

static VAStatus va_TranslateBufferCompat (
    VADriverContextP ctx,
    VABufferCompat *compat_buffer
)
{
    VAStatus status;
    unsigned char *src, *dest;
    int i;

    status = VA_INVOKE(vaMapBuffer, ctx, compat_buffer->id, (void **)&src);
    if (status != VA_STATUS_SUCCESS)
        return status;

    ASSERT(compat_buffer->data);
    ASSERT(src == compat_buffer->data);
    ASSERT(compat_buffer->compat_data);
    dest = compat_buffer->compat_data;
    for (i = 0; i < compat_buffer->num_elements; i++)
    {
        /* XXX: check errors */
        status = compat_buffer->translate(ctx->compat, dest, src);

        src   += compat_buffer->size;
        dest  += compat_buffer->compat_size;
    }

    return VA_INVOKE(vaUnmapBuffer, ctx, compat_buffer->id);
}

static VAStatus va_CreateBufferCompat (
    VADriverContextP ctx,
    VAContextID context,
    VABufferID id,
    VABufferType type,
    unsigned int size,
    unsigned int num_elements,
    unsigned int compat_size,
    VABufferTranslateFunc translate_func
)
{
    VACompatContextP compat = ctx->compat;
    VABufferCompat *compat_buffer;
    int index;

    if (!va_IsIntelBuffer(ctx, id))
        return VA_STATUS_ERROR_INVALID_BUFFER;
    index = id & 0x00ffffff;

    /* XXX: this allocation strategy is not really space efficient... */
    if (index >= compat->buffers_count_max)
    {
        compat->buffers_count_max = index + 1;
        compat->buffers = realloc(compat->buffers,
                                  (compat->buffers_count_max *
                                   sizeof(VABufferCompat)));
        if (compat->buffers == NULL)
            return VA_STATUS_ERROR_ALLOCATION_FAILED;
    }

    compat_buffer = &compat->buffers[index];
    compat_buffer->id           = id;
    compat_buffer->type         = type;
    compat_buffer->size         = size;
    compat_buffer->num_elements = num_elements;
    compat_buffer->data         = NULL;
    compat_buffer->translate    = translate_func;
    compat_buffer->compat_size  = compat_size;
    compat_buffer->compat_data  = NULL;
    compat_buffer->map_count    = 0;

    return VA_INVOKE_COMPAT(vaCreateBuffer,
                            ctx,
                            context,
                            type,
                            compat_size,
                            num_elements,
                            NULL,
                            &compat_buffer->compat_id);
}

static VAStatus va_DestroyBufferCompat (
    VADriverContextP ctx,
    VABufferID id
)
{
    VABufferCompat *compat_buffer;
    VAStatus status;

    if ((compat_buffer = va_GetBufferCompat(ctx, id)) == NULL)
        return VA_STATUS_SUCCESS;

    /* Force unmap if there were more maps than unmaps */
    if (compat_buffer->map_count > 1)
        compat_buffer->map_count = 1;
    if (compat_buffer->map_count > 1)
    {
        if ((status = VA_INVOKE(vaUnmapBuffer, ctx, id)) != VA_STATUS_SUCCESS)
            return status;
    }

    compat_buffer->id = 0;
    return VA_INVOKE_COMPAT(vaDestroyBuffer, ctx, compat_buffer->compat_id);
}

static VAStatus va_compat_CreateBuffer (
    VADriverContextP ctx,
    VAContextID context,	/* in */
    VABufferType type,		/* in */
    unsigned int size,		/* in */
    unsigned int num_elements,	/* in */
    void *data,			/* in */
    VABufferID *buf_id		/* out */
)
{
    VABufferTranslateFunc translate_func = NULL;
    unsigned int compat_size = 0;
    VAStatus status;

    status = VA_INVOKE_COMPAT(vaCreateBuffer,
                              ctx,
                              context,
                              type,
                              size,
                              num_elements,
                              data,
                              buf_id);

    if (status != VA_STATUS_SUCCESS)
        return status;

    /* XXX: this assumes all data structures have different sizes */
    switch (size) {
    case sizeof(VAPictureParameterBufferH264):
        if (type == VAPictureParameterBufferType)
        {
            translate_func = va_compat_TranslateVAPictureParameterBufferH264;
            compat_size = sizeof(VAPictureParameterBufferH264_0_29);
        }
        break;
    case sizeof(VASliceParameterBufferH264):
        if (type == VASliceParameterBufferType)
        {
            translate_func = va_compat_TranslateVASliceParameterBufferH264;
            compat_size = sizeof(VASliceParameterBufferH264_0_29);
        }
        break;
    case sizeof(VAPictureParameterBufferVC1):
        if (type == VAPictureParameterBufferType)
        {
            translate_func = va_compat_TranslateVAPictureParameterBufferVC1;
            compat_size = sizeof(VAPictureParameterBufferVC1_0_29);
        }
        break;
    case sizeof(VAPictureParameterBufferMPEG2):
        if (type == VAPictureParameterBufferType)
        {
            translate_func = va_compat_TranslateVAPictureParameterBufferMPEG2;
            compat_size = sizeof(VAPictureParameterBufferMPEG2_0_29);
        }
        break;
    case sizeof(VAPictureParameterBufferMPEG4):
        if (type == VAPictureParameterBufferType)
        {
            translate_func = va_compat_TranslateVAPictureParameterBufferMPEG4;
            compat_size = sizeof(VAPictureParameterBufferMPEG4_0_29);
        }
        break;
    }

    /* Create thunk */
    if (buf_id && translate_func)
    {
        ASSERT(compat_size > 0);

        status = va_CreateBufferCompat(ctx,
                                       context,
                                       *buf_id,
                                       type,
                                       size,
                                       num_elements,
                                       compat_size,
                                       translate_func);
    }

    return status;
}

static VAStatus va_compat_BufferSetNumElements (
    VADriverContextP ctx,
    VABufferID buf_id,	/* in */
    unsigned int num_elements /* in */
)
{
    VABufferCompat *compat_buffer;
    VAStatus status;

    status = VA_INVOKE_COMPAT(vaBufferSetNumElements, ctx, buf_id, num_elements);
    if (status != VA_STATUS_SUCCESS)
        return status;

    if ((compat_buffer = va_GetBufferCompat(ctx, buf_id)) != NULL)
    {
        compat_buffer->num_elements = num_elements;
        status = VA_INVOKE_COMPAT(vaBufferSetNumElements,
                                  ctx,
                                  compat_buffer->compat_id,
                                  num_elements);
    }

    return status;
}

static VAStatus va_compat_DestroyBuffer (
    VADriverContextP ctx,
    VABufferID buffer_id
)
{
    VAStatus status;
    if ((status = va_DestroyBufferCompat(ctx, buffer_id)) != VA_STATUS_SUCCESS)
        return status;

    return VA_INVOKE_COMPAT(vaDestroyBuffer, ctx, buffer_id);
}

static VAStatus va_compat_MapBuffer (
    VADriverContextP ctx,
    VABufferID buf_id,	/* in */
    void **pbuf 	/* out */
)
{
    VABufferCompat *compat_buffer;
    VAStatus status;

    if ((compat_buffer = va_GetBufferCompat(ctx, buf_id)) == NULL)
        return VA_INVOKE_COMPAT(vaMapBuffer, ctx, buf_id, pbuf);

    if (compat_buffer->map_count++ == 0)
    {
        status = VA_INVOKE_COMPAT(vaMapBuffer, ctx, buf_id, (void **)&compat_buffer->data);
        if (status != VA_STATUS_SUCCESS)
            return status;

        status = VA_INVOKE_COMPAT(vaMapBuffer, ctx, compat_buffer->compat_id, (void **)&compat_buffer->compat_data);
        if (status != VA_STATUS_SUCCESS)
            return status;
    }

    if (pbuf)
        *pbuf = compat_buffer->data;

    return VA_STATUS_SUCCESS;
}

static VAStatus va_compat_UnmapBuffer (
    VADriverContextP ctx,
    VABufferID buf_id	/* in */
)
{
    VABufferCompat *compat_buffer;
    VAStatus status;

    if ((compat_buffer = va_GetBufferCompat(ctx, buf_id)) == NULL)
        return VA_INVOKE_COMPAT(vaUnmapBuffer, ctx, buf_id);

    if (--compat_buffer->map_count == 0)
    {
        status = VA_INVOKE_COMPAT(vaUnmapBuffer, ctx, compat_buffer->compat_id);
        if (status != VA_STATUS_SUCCESS)
            return status;
        compat_buffer->compat_data = NULL;

        status = VA_INVOKE_COMPAT(vaUnmapBuffer, ctx, compat_buffer->id);
        if (status != VA_STATUS_SUCCESS)
            return status;
        compat_buffer->data = NULL;
    }

    return VA_STATUS_SUCCESS;
}

static VAStatus va_compat_BeginPicture (
    VADriverContextP ctx,
    VAContextID context,
    VASurfaceID render_target
)
{
    VACompatContextP compat = ctx->compat;
    compat->skip_frame = 0;
    return VA_INVOKE_COMPAT(vaBeginPicture, ctx, context, render_target);
}

static VAStatus va_compat_EndPicture (
    VADriverContextP ctx,
    VAContextID context
)
{
    VACompatContextP compat = ctx->compat;
    VAStatus status = VA_INVOKE_COMPAT(vaEndPicture, ctx, context);

    /* Ignore errors if the HW decoder did not handle VC-1 skipped P-frames */
    if (compat->skip_frame && status == VA_STATUS_ERROR_UNKNOWN)
        status = VA_STATUS_SUCCESS;

    return status;
}

static VAStatus va_compat_RenderPicture (
    VADriverContextP ctx,
    VAContextID context,
    VABufferID *buffers,
    int num_buffers
)
{
    VACompatContextP compat = ctx->compat;
    VABufferCompat *compat_buffer;
    VABufferID *compat_buffer_ids;
    VAStatus status;
    int i, n;

    if ((n = num_buffers) < 1)
        n = 1;
    compat_buffer_ids = alloca(n * sizeof(compat_buffer_ids[0]));

    for (i = 0; i < num_buffers; i++)
    {
        if ((compat_buffer = va_GetBufferCompat(ctx, buffers[i])) == NULL)
            compat_buffer_ids[i] = buffers[i];
        else
        {
            status = va_TranslateBufferCompat(ctx, compat_buffer);
            if (status != VA_STATUS_SUCCESS)
                return status;
            compat_buffer_ids[i] = compat_buffer->compat_id;
        }
    }

    if (!compat->skip_frame)
    {
        status = VA_INVOKE_COMPAT(vaRenderPicture, ctx, context, compat_buffer_ids, num_buffers);
        if (status != VA_STATUS_SUCCESS)
            return status;
    }

    /* Buffers are automatically destroyed afterwards */
    for (i = 0; i < num_buffers; i++)
    {
        if ((compat_buffer = va_GetBufferCompat(ctx, buffers[i])) != NULL)
        {
            status = VA_INVOKE_COMPAT(vaDestroyBuffer, ctx, compat_buffer->id);
            if (status != VA_STATUS_SUCCESS)
                return status;
        }
    }

    return VA_STATUS_SUCCESS;
}

VAStatus va_compat_init(VADisplay dpy, int is_compat_needed)
{
    VADisplayContextP pDisplayContext = (VADisplayContextP)dpy;
    VADriverContextP ctx = CTX(dpy);
    VACompatContextP compat;
    VAStatus status;

    ctx->compat                         = NULL;

    if (!is_compat_needed)
        return VA_STATUS_SUCCESS;

    if ((compat = malloc(sizeof(*compat))) == NULL)
        return VA_STATUS_ERROR_ALLOCATION_FAILED;
    compat->buffers                     = NULL;
    compat->buffers_count_max           = 0;
    ctx->compat                         = compat;

    status = pDisplayContext->vaGetDriverName(pDisplayContext,
                                              &compat->driver_name);
    ASSERT(status == VA_STATUS_SUCCESS);
    if (status != VA_STATUS_SUCCESS)
        return status;

    if (strcmp(compat->driver_name, "psb") == 0)
        compat->driver_id = VA_DRIVER_ID_POULSBO;
    else if (strcmp(compat->driver_name, "iegd") == 0)
        compat->driver_id = VA_DRIVER_ID_IEGD;
    else
        compat->driver_id = VA_DRIVER_ID_UNKNOWN;

    memcpy(&compat->vtable, &ctx->vtable, sizeof(compat->vtable));
    ctx->vtable.vaCreateBuffer          = va_compat_CreateBuffer;
    ctx->vtable.vaDestroyBuffer         = va_compat_DestroyBuffer;
    ctx->vtable.vaBufferSetNumElements  = va_compat_BufferSetNumElements;
    ctx->vtable.vaMapBuffer             = va_compat_MapBuffer;
    ctx->vtable.vaUnmapBuffer           = va_compat_UnmapBuffer;
    ctx->vtable.vaBeginPicture          = va_compat_BeginPicture;
    ctx->vtable.vaEndPicture            = va_compat_EndPicture;
    ctx->vtable.vaRenderPicture         = va_compat_RenderPicture;
    return VA_STATUS_SUCCESS;
}

VAStatus va_compat_fini(VADisplay dpy)
{
    VADriverContextP ctx = CTX(dpy);
    VACompatContextP compat = ctx->compat;
    int i;

    if (compat == NULL)
        return VA_STATUS_SUCCESS;

    if (compat->driver_name)
    {
        free(compat->driver_name);
        compat->driver_name = NULL;
    }

    if (compat->buffers)
    {
        for (i = 0; i < compat->buffers_count_max; i++)
        {
            if (compat->buffers[i].id)
                va_DestroyBufferCompat(ctx, compat->buffers[i].id);
        }
        free(compat->buffers);
        compat->buffers = NULL;
    }

    free(compat);
    ctx->compat = NULL;
    return VA_STATUS_SUCCESS;
}
